# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import logging
import os
import requests
import subprocess
import sys
import tempfile
import threading
import time
import zipfile

import config
import log
import paths

UNZIP_SUBPROCESS_LINE = [sys.executable, "-m", "zipfile", "-e"]

class Downloader(threading.Thread):

    CONNECTION_TIMEOUT = 6
    MAX_RETRIES = 15

    def __init__(self, parent, url, callback, is_zip):
        self.percent = 0
        self.url = url
        self.is_zip = is_zip
        self.parent = parent
        self.callback = callback
        self.cancel_flag = False
        self.logger = logging.getLogger(__name__)
        threading.Thread.__init__(self, name="Downloader", daemon=True)

    @log.log_exception
    def run(self):
        self.logger.info('Starting downloading')
        downloaded_filename = self.download()
        if downloaded_filename:
            try:
                if self.is_zip:
                    if config.get_settings()['subprocess_zip']:
                        with tempfile.TemporaryDirectory(dir=paths.DOWNLOAD_FOLDER) as unzip_tmpdir_name:
                            proc_args = UNZIP_SUBPROCESS_LINE + [downloaded_filename, unzip_tmpdir_name]
                            proc = subprocess.Popen(proc_args)
                            while not self.parent or not self.parent.stop_flag:
                                exit_code = proc.poll()
                                if exit_code == None:
                                    time.sleep(0.1)
                                elif proc and proc.poll() == 0:
                                    filename = os.path.join(unzip_tmpdir_name, os.listdir(unzip_tmpdir_name)[0])
                                    with open(filename, 'rb') as f:
                                        self.execule_callback(f)
                                    os.remove(filename)
                                    break
                                else:
                                    self.logger.warning(f'Unable to unzip {downloaded_filename}')
                                    raise zipfile.BadZipFile
                    else:
                        with zipfile.ZipFile(downloaded_filename) as zf:
                            filename = zf.namelist()[0]
                            with zf.open(filename, 'r') as f:
                                self.execule_callback(f)
                else:
                    with open(downloaded_filename, 'rb') as f:
                        self.execule_callback(f)
                self.logger.info('Gcodes loaded to memory, deleting temp file')
                os.remove(downloaded_filename)
            except IndexError:
                self.parent.register_error(86, "Empty zipfile. Cancelling...", is_blocking=True)
            except (OSError, IOError):
                self.parent.register_error(87, "Temporary file error. Cancelling...", is_blocking=True)
            except MemoryError:
                self.parent.register_error(88, "Out of memory. Cancelling...", is_blocking=True)
            except zipfile.error:
                self.parent.register_error(89, "Bad zip file. Cancelling...", is_blocking=True)
        if self.cancel_flag:
            self.logger.info('Cancel command was received after printing start in downloading thread')
            try:
                self.parent.printer.cancel()
            except AttributeError:
                pass
        self.logger.info('Downloading finished')

    def execule_callback(self, f):
        if self.cancel_flag:
            self.logger.info('Cancel command received')
            self.cancel_flag = False
        elif not f:
            self.parent.register_error(67, "Unknown download error", is_blocking=True)
        else:
            self.callback(f)

    def download(self):
        self.logger.info("Downloading from " + self.url)
        if self.is_zip:
            suffix = ".zip"
        else:
            suffix = ".g"
        self.tmp_file = tempfile.NamedTemporaryFile(mode='wb', dir=paths.DOWNLOAD_FOLDER,
                delete=False, prefix='3dprinteros-', suffix=suffix)
        filename = self.tmp_file.name
        resume_byte_pos = 0
        retry = 0
        while retry < self.MAX_RETRIES and not self.parent.stop_flag and not self.cancel_flag:
            if retry:
                self.logger.warning("Download retry/resume N%d" % retry)
            self.logger.info("Connecting to " + self.url)
            resume_header = {'Range': 'bytes=%d-' % resume_byte_pos}
            try:
                request = requests.get(self.url, headers = resume_header, stream=True, timeout = self.CONNECTION_TIMEOUT)
            except Exception as e:
                request = None
                self.parent.register_error(65, "Unable to open download link: " + str(e), is_blocking=False)
            else:
                self.logger.info("Successful connection to " + self.url)
                download_length = int(request.headers.get('content-length', 0))
                if download_length:
                    downloaded_size = self.download_chunks(request, download_length)
                    resume_byte_pos += downloaded_size
                    self.logger.info("Downloaded %d bytes" % resume_byte_pos)
                    if downloaded_size == download_length:
                        self.tmp_file.close()
                        return filename
            finally:
                if request:
                    request.close()
                retry += 1
                time.sleep(1)
        self.tmp_file.close()
        try:
            os.remove(filename)
        except:
            pass

    def download_chunks(self, request, download_length):
        # Taking +1 byte with each chunk to compensate file length till less than 100 bytes when dividing by 100
        percent_length = download_length // 100 + 1
        total_size = 0
        try:
            for chunk in request.iter_content(percent_length):
                if self.cancel_flag or self.parent.stop_flag:
                    self.logger.info('Download canceled')
                    break
                self.tmp_file.write(chunk)
                self.percent += 1
                total_size += len(chunk)
                self.logger.info('File downloading : %d%%' % self.percent)
        except Exception as e:
            self.parent.register_error(66, 'Error while downloading: ' + str(e), is_blocking=False)
        finally:
            return total_size

    def cancel(self):
        self.cancel_flag = True

    def get_percent(self):
        return self.percent
